@@ -136,18 +136,26 @@ class ProjectObj(object):
     def validate_label_config(cls, config_string):
         # xml and schema
         try:
+            logger.debug('Convert label config from XML to JSON')
             config = cls.parse_config_to_json(config_string)
+            logger.debug(json.dumps(dict(config), indent=2))
             jsonschema.validate(config, _LABEL_CONFIG_SCHEMA_DATA)
         except (etree.XMLSyntaxError, etree.XMLSchemaParseError, ValueError) as exc:
+            logger.debug('Parsing error')
             raise ValidationError(str(exc))
         except jsonschema.exceptions.ValidationError as exc:
+            logger.debug('Validation error')
             error_message = exc.context[-1].message if len(exc.context) else exc.message
             error_message = 'Validation failed on {}: {}'.format('/'.join(exc.path), error_message.replace('@', ''))
             raise ValidationError(error_message)
+        except Exception as exc:
+            logger.debug('Unknown error: ' + str(exc))
+            raise ValidationError(str(exc))
 
         # unique names in config # FIXME: 'name =' (with spaces) won't work
         all_names = re.findall(r'name="([^"]*)"', config_string)
         if len(set(all_names)) != len(all_names):
+            logger.debug(all_names)
             raise ValidationError('Label config contains non-unique names')
 
         # toName points to existent name
@@ -250,9 +258,10 @@ class MLApiScheme(object):
 
 class MLApi(BaseHTTPAPI):
 
-    def __init__(self, url, **kwargs):
+    def __init__(self, url, name, **kwargs):
         super(MLApi, self).__init__(url=url, **kwargs)
         self._validate_request_timeout = 10
+        self._name = name
 
     def is_ok(self):
         url_is_ok = self._url is not None and isinstance(self._url, str)
@@ -264,6 +273,30 @@ class MLApi(BaseHTTPAPI):
             url += '/'
         return urllib.parse.urljoin(url, url_suffix)
 
+    def _get(self, url_suffix, *args, **kwargs):
+        url = self._get_url(url_suffix)
+        headers = dict(self.http.headers)
+        response = None
+        try:
+            response = self.get(url=url, *args, **kwargs)
+            response.raise_for_status()
+        except requests.exceptions.RequestException as e:
+            logger.debug('Error getting response from ' + url, exc_info=True)
+            status_code = response.status_code if response is not None else 0
+            error_msg = response.content.decode('utf-8') if response is not None else str(e)
+            return MLApiResult(url, {}, {'error': error_msg}, headers, 'error', status_code=status_code)
+        status_code = response.status_code
+        try:
+            response = response.json()
+        except ValueError as e:
+            logger.debug('Error parsing JSON response from '+url+'. Response: ' + str(response.content), exc_info=True)
+            return MLApiResult(
+                url, {}, {'error': str(e), 'response': response.content.decode('utf-8')}, headers, 'error',
+                status_code=status_code
+            )
+        logger.debug('Response from ' + url + ':' + json.dumps(response, indent=2))
+        return MLApiResult(url, {}, response, headers, status_code=status_code)
+
     def _post(self, url_suffix, request, *args, **kwargs):
         url = self._get_url(url_suffix)
         headers = dict(self.http.headers)
@@ -275,7 +308,8 @@ class MLApi(BaseHTTPAPI):
         except requests.exceptions.RequestException as e:
             logger.debug('Error getting response from ' + url, exc_info=True)
             status_code = response.status_code if response is not None else 0
-            return MLApiResult(url, request, {'error': str(e)}, headers, 'error', status_code=status_code)
+            error_msg = response.content.decode('utf-8') if response is not None else str(e)
+            return MLApiResult(url, request, {'error': error_msg}, headers, 'error', status_code=status_code)
         status_code = response.status_code
         try:
             response = response.json()
@@ -288,34 +322,21 @@ class MLApi(BaseHTTPAPI):
         logger.debug('Response from ' + url + ':' + json.dumps(response, indent=2))
         return MLApiResult(url, request, response, headers, status_code=status_code)
 
-    @staticmethod
-    def _create_project_uid(project):
-        return str(project.id) + '.' + project.ml_backend.model_name
-
-    def update(self, task, results, project, retrain=True):
-        """
-        Upload new task results and update model when necessary
-        :param task:
-        :param results:
-        :param project:
-        :param retrain:
-        :return:
-        """
+    def _create_project_uid(self, project):
+        return self._name
 
+    def train(self, completions, project):
+        """Upload new task results and update model when necessary"""
         request = {
-            'id': task['id'],
+            'completions': completions,
             'project': self._create_project_uid(project),
-            'schema': project.schema,
-            'data': task['data'],
-            'meta': task.get('meta', {}),
-            'result': results,
-            'retrain': retrain,
+            'label_config': project.label_config_line,
             'params': {
                 'login': project.task_data_login,
                 'password': project.task_data_password
             }
         }
-        return self._post('update', request)
+        return self._post('train', request)
 
     def predict(self, tasks, model_version, project):
         """
@@ -329,8 +350,9 @@ class MLApi(BaseHTTPAPI):
             'tasks': tasks,
             'model_version': model_version,
             'project': self._create_project_uid(project),
-            'schema': project.schema,
+            'label_config': project.label_config_line,
             'params': {
+                'force_load': True,
                 'login': project.task_data_login,
                 'password': project.task_data_password
             }
@@ -353,7 +375,7 @@ class MLApi(BaseHTTPAPI):
         """
         return self._post('setup', request={
             'project': self._create_project_uid(project),
-            'schema': project.schema
+            'schema': project.label_config_line
         })
 
     def delete(self, project):
@@ -372,6 +394,20 @@ class MLApi(BaseHTTPAPI):
         """
         return self._post('job_status', request={'job': train_job})
 
+    def is_training(self, project):
+        return self._get('is_training?project=' + self._create_project_uid(project))
+
+    def check_connection(self):
+        return self._get('health')
+
+
+class CantStartTrainJobError(Exception):
+    pass
+
+
+class CantValidateIsTraining(Exception):
+    pass
+
 
 @attr.s
 class MLBackend(object):
@@ -386,9 +422,35 @@ class MLBackend(object):
     model_version = attr.ib(default=None)
     # train job running on ML backend
     train_job = attr.ib(default=None)
+    # number of completions fed
+    num_completions = attr.ib(type=int, default=0)
+    # backend dir
+    _dir = attr.ib(default=None)
 
     _TRAIN_JOBS_FILE = os.path.join(get_data_dir(), 'train_jobs.json')
 
+    @property
+    def url(self):
+        return self.api._url
+
+    @property
+    def dir(self):
+        if self._dir is None:
+            r = self.api.check_connection()
+            if not r.is_error:
+                self._dir = os.path.basename(r.response['model_dir'])
+        return self._dir
+
+    @property
+    def connected(self):
+        try:
+            r = self.api.check_connection()
+            if r.is_error:
+                return False
+        except:
+            return False
+        return True
+
     def restore_train_job(self):
         """
         Restore train job for the given model name
@@ -403,9 +465,9 @@ class MLBackend(object):
                 logger.warning(
                     'Can\'t restore train job because ' + self.model_name + ' key not found in ' + self._TRAIN_JOBS_FILE)
             else:
-                train_job = train_jobs[self.model_name]
+                self.train_job = train_jobs[self.model_name]
                 logger.debug(
-                    'Train job ' + train_job + ' for model name ' + self.model_name +
+                    'Train job ' + self.train_job + ' for model name ' + self.model_name +
                     ' restored from ' + self._TRAIN_JOBS_FILE)
 
     def save_train_job(self):
@@ -423,17 +485,17 @@ class MLBackend(object):
 
     @classmethod
     def from_params(cls, params):
-        ml_api = MLApi(params['url'])
+        ml_api = MLApi(params['url'], params['name'])
         m = MLBackend(api=ml_api, model_name=params['name'])
         m.restore_train_job()
         return m
 
-    def train_job_is_running(self, project):
-        if self._api_exists() and project.train_job is not None:
-            response = self.api.get_train_job_status(project.train_job)
+    def train_job_is_running(self):
+        if self._api_exists() and self.train_job is not None:
+            response = self.api.get_train_job_status(self.train_job)
             if response.is_error:
-                logger.error('Can\'t fetch train job status for job ' + project.train_job + ': '
-                             'ML backend returns error: ' + response.error_message)
+                logger.error('Can\'t fetch train job status for job ' + self.train_job + ': '
+                             'ML backend returns an error: ' + response.error_message)
             else:
                 return response.response['job_status'] in ('queued', 'started')
         return False
@@ -445,40 +507,52 @@ class MLBackend(object):
             return False
         return True
 
-    def make_predictions(self, task, project):
-        if self._api_exists():
-            r = self.api.setup(project)
-            if not r.is_error:
-                model_version = r.response['model_version']
-                if self.model_version != model_version:
-                    self.model_version = model_version
-                    logger.debug('Model version has changed: ' + model_version)
-                else:
-                    logger.debug('Model version hasn\'t changed: ' + model_version)
-                response = self.api.predict([task], self.model_version, project)
-                if response.is_error:
-                    if response.status_code == 404:
-                        logger.info('Can\'t make predictions: model is not found (probably not trained yet)')
-                    else:
-                        logger.error('Can\'t make predictions: ML backend returns error: ' + response.error_message)
-                else:
-                    return response.response['results']
-            else:
-                logger.warning('ML backend setup returns error: ' + r.error_message)
+    def sync(self, project):
+        r = self.api.setup(project)
+        if r.is_error:
+            raise ValueError(r.error_message)
+        model_version = r.response['model_version']
+        if self.model_version != model_version:
+            self.model_version = model_version
+            logger.debug('Model version has changed: ' + str(model_version))
+        else:
+            logger.debug('Model version hasn\'t changed: ' + str(model_version))
 
-    def update_model(self, task, completion, project):
-        if self._api_exists():
-            results = completion['result']
-            retrain = not self.train_job_is_running(project)
-            response = self.api.update(task, results, project, retrain)
-            if response.is_error:
-                logger.error('Can\'t update model: ML backend returns error: ' + response.error_message)
+    def make_predictions(self, task, project):
+        self.sync(project)
+        response = self.api.predict([task], self.model_version, project)
+        if response.is_error:
+            if response.status_code == 404:
+                logger.info('Can\'t make predictions: model is not found (probably not trained yet)')
             else:
-                maybe_job = response.response.get('job')
-                if maybe_job:
-                    self.train_job = maybe_job
-                    self.save_train_job()
-                    logger.debug('Project ' + str(project) + ' successfully updated train job ' + self.train_job)
+                logger.error('Can\'t make predictions: ML backend returns an error: ' + response.error_message)
+        else:
+            return response.response['results'][0]
+
+    def is_training(self, project):
+        if self._api_exists():
+            response = self.api.is_training(project)
+            if response.is_error:
+                raise CantValidateIsTraining('Can\'t validate whether model is training for project ' + project.name)
+            return response.response
+
+    def train(self, completions, project):
+        if self.train_job_is_running():
+            raise CantStartTrainJobError('Can\'t start new training: Train job is running.')
+        train_status = self.is_training(project)
+        if train_status['is_training']:
+            raise CantStartTrainJobError('Can\'t start new training: Train job is running.')
+        response = self.api.train(completions, project)
+        if response.is_error:
+            raise CantStartTrainJobError('Can\'t update model: ML backend returns an error: ' + response.error_message)
+        else:
+            self.num_completions = len(completions)
+            logger.info('Training job with ' + str(self.num_completions) + ' completions has been started.')
+            maybe_job = response.response.get('job')
+            if maybe_job:
+                self.train_job = maybe_job
+                self.save_train_job()
+                logger.debug('Project ' + str(project) + ' successfully updated train job ' + self.train_job)
 
     def validate(self, label_config):
         if self._api_exists():
@@ -510,3 +584,9 @@ class MLBackend(object):
                 logger.error('Can\'t clear ML backend for project ' + project.name + ': ' + response.error_message)
             else:
                 logger.info('ML backend for project ' + project.name + ' has been cleared.')
+
+    def train_log(self):
+        return 'here is train log\n' * 100
+
+    def prediction_log(self):
+        return 'here is prediction log\n' * 100
